import std/[algorithm, logging, nre, options, os, random, strutils]
import uuids, zero_functional
import ./config, ./db, ./library, ./libvlc, ./models, ./whenmatched
import private/cliconstants, private/styledtext

type

  Selection = ref object
    models*: seq[BaseModel]

  CliCtx = ref object
    cfg: WdiwtltConfig
    library: WdiwtltLibrary
    player: VlcMediaPlayer
    vlc: LibVlcInstance

    currentBookmark: Option[Bookmark]
    currentSelection: Option[Selection]
    playQueue: Option[Playlist]
    curMediaFile: Option[MediaFile]

const STOP_CMDS = ["stop", "exit", "quit", ":q"]
var cmdChannel: Channel[string]
var readyForCmd: Channel[bool]

proc initSelection(models: seq[SomeModel]): Selection =
  Selection(models: models --> map(it.BaseModel))

proc handleCmd(ctx: CliCtx, cmd: string): bool;

proc err(msg: string) = raise newException(CliErr, msg)

proc initCliCtx(cfg: WdiwtltConfig): CliCtx =
  let vlc = newVlc()
  randomize()
  return CliCtx(
    cfg: cfg,
    player: vlc.newMediaPlayer,
    vlc: vlc,
    library: initLibrary(cfg.libraryPath, cfg.dbPath, vlc))

proc release(ctx: CliCtx) =
  if not ctx.player.isNil: ctx.player.release
  if not ctx.vlc.isNil: ctx.vlc.release

proc readInputLoop() =
  var line: string
  while not STOP_CMDS.contains(line):
    stdout.write("> ")
    line = stdin.readline
    cmdChannel.send(line)

    while not readyForCmd.tryRecv().dataAvailable: sleep(200)

proc renderLoop(ctx: CliCtx) =
  var shouldStop = false
  while not shouldStop:
    # poll for input
    let cmdMsg = cmdChannel.tryRecv()
    if cmdMsg.dataAvailable:
      shouldStop = ctx.handleCmd(cmdMsg.msg)
      readyForCmd.send(true)

    # render
    sleep(200)

proc startCli*(cfg: WdiwtltConfig) =
  var ctx = cfg.initCliCtx

  try:
    stdout.writeLine("wdiwtlt v" & WDIWTLT_VERSION)
    cmdChannel.open()
    readyForCmd.open()

    var readLineWorker: Thread[void]
    createThread(readLineWorker, readInputLoop)
    ctx.renderLoop()
    readLineWorker.joinThread()

  finally:
    readyForCmd.close()
    cmdChannel.close()
    ctx.release()

proc processScan(ctx: CliCtx, fullRescan = false) =
  stdout.writeLine("Sanning media library...")
  ctx.library.scan(fullRescan)

const SELECTABLE_MODELS = "album|artist|file|playlist|tag"

proc selectFrom(ctx: CliCtx, mf: MediaFile, modelType: ModelType): Selection =
  let db = ctx.library.db
  case modelType:
  of mtAlbum: result = initSelection(db.findAlbumsByMediaFile(mf))
  of mtArtist: result = initSelection(db.findArtistsByMediaFile(mf))
  of mtBookmark:
    if ctx.currentBookmark.isSome:
      result = initSelection(@[ctx.currentBookmark.get])
    else: err "no current bookmark to select"

  of mtMediaFile: result = initSelection(@[mf])
  of mtPlaylist:
    if ctx.playQueue.isSome: result = initSelection(@[ctx.playQueue.get])
    else: err "no current playlist to select"
  of mtTag: result = initSelection(db.findTagsByMediaFile(mf))

proc selectRandom(
    ctx: CliCtx,
    count: int,
    source: Selection,
    mt: ModelType
  ): Selection =

  var options = ctx.library.db.findAllRelatedModels(source.models, mt)
  if count >= options.len: return initSelection(options)

  shuffle(options)
  return initSelection(options[0..<count])

proc select(ctx: CliCtx, criteria: string): Selection =
  debug "Selecting. Options: " & criteria
  let db = ctx.library.db

  # playing
  criteria.whenMatched(re("playing ($#)" % SELECTABLE_MODELS)):
    if ctx.curMediaFile.isNone: err "Nothing selected."

    return ctx.selectFrom(
      ctx.curMediaFile.get,
      mapModelType(match.captures[0]))

  # <count> random <model-type>s from <selection-criteria>
  criteria.whenMatched(re(
    "(\\d+ )?random ($#)s?( from (.+$))?" % SELECTABLE_MODELS)):

    let count =
      if match.captures.contains(0): parseInt(match.captures[0].strip)
      else: 1

    let modelType = mapModelType(match.captures[1])
    let source: Selection =
      if match.captures.contains(3): ctx.select(match.captures[3].strip)
      else: initSelection(getAll(db, modelType))

    return ctx.selectRandom(count, source, modelType)

  # selected <model-ype>
  criteria.whenMatched(re("selected ($#)s?" % SELECTABLE_MODELS)):
    if ctx.currentSelection.isNone: err "Nothing is selected."
    let modelType = mapModelType(match.captures[0])

    return initSelection(
      db.findAllRelatedModels(ctx.currentSelection.get.models, modelType))

  # absent files
  criteria.whenMatched(re("absent files")):
    return initSelection(db.allMediaFiles --> filter(not it.presentLocally))

  # files tagged as <tags>...
  criteria.whenMatched(re("files tagged( as){0,1}((\\s[^\\s]+)+)")):

    let tags: seq[Tag] = match.captures[1].split(re"\s") -->
      map(it.strip).
      map(db.findTagByName(it)).
      filter(it.isSome).
      map(it.get)

    if tags.len == 0: err "Nothing is selected."
    return initSelection(tags --> map(db.findMediaFilesByTag(it)).flatten())

  # <model-type> from <selection-criteria>
  criteria.whenMatched(re("($#)s? from (.+)" % SELECTABLE_MODELS)):
    let modelType = mapModelType(match.captures[0])
    let source = ctx.select(match.captures[1])

    if source.models.len > 0 and modelMatchesType(source.models[0], modelType):
      return source

    return initSelection(db.findAllRelatedModels(source.models, modelType))

  # <model-type> <id>
  #criteria.whenMatched(re("($#)s?((\\s\\[a-fA-F0-9]+)+)" % SELECTABLE_MODELS)):
  #  let modelType = mapModelType(match.captures[0])
  #  let ids = match.captures[1].split(re"\s") --> map(parseUUID(it))

proc processList(ctx: CliCtx, options: string) =
  debug "Listing things. Options: " & options

proc processHelp(ctx: CliCtx, options: string) =
  stdout.writeStyledText(
    case options:
    of "": HELP_GENERAL
    of "scan": HELP_SCAN
    of "rescan": HELP_RESCAN
    of "quit": HELP_QUIT
    else: FE & options & " is not a recognized command.\p"
  )

proc handleCmd(ctx: CliCtx, cmd: string): bool =
  try:
    result = false
    if STOP_CMDS.contains(cmd): return true

    let cmdParts = cmd.split(" ", 1)
    if cmdParts.len == 0: return
    let command = cmdParts[0].toLower
    let rest =
      if cmdParts.len > 1: cmdParts[1]
      else: ""

    case command:
      of "scan": ctx.processScan()
      of "rescan": ctx.processScan(true)
      of "list": ctx.processList(rest)
      of "help": ctx.processHelp(rest)
      else: stdout.writeLine("Unrecognized command: '" & cmdParts[0] & "'")
  except:
    stdout.writeStyledText(FE & getCurrentExceptionMsg() & FN)
