import std/[logging, json, options, os, sets, tables, times]
import timeutils, uuids, zero_functional

import ./jsonutils, ./models

from std/sequtils import deduplicate, toSeq

type
  DbRoot = ref object
    albums: TableRef[UUID, Album]
    artists: TableRef[UUID, Artist]
    bookmarks: TableRef[UUID, Bookmark]
    mediaFiles: TableRef[UUID, MediaFile]
    playlists: TableRef[UUID, Playlist]
    tags: TableRef[UUID, Tag]
    tagSet: HashSet[string]

    albumsToMediaFiles: TableRef[UUID, seq[UUID]]
    artistsToMediaFiles: TableRef[UUID, seq[UUID]]
    artistsAndAlbums: seq[tuple[artistId, albumId: UUID]]
    playlistsToMediaFiles: TableRef[UUID, seq[UUID]]
    tagsAndMediaFiles: seq[tuple[tagId: UUID, mediaFileId: UUID]]

    mediaFileHashToId: TableRef[string, UUID]

  WdiwtltDb* = ref object
    jsonFilePath: string
    root: DbRoot

proc debug*(db: WdiwtltDb): string =
  "\p\talbums: " & $db.root.albums &
  "\p\tartists: " & $db.root.artists &
  "\p\tbookmarks: " & $db.root.bookmarks &
  "\p\tmediaFiles: " & $db.root.mediaFiles &
  "\p\tplaylists: " & $db.root.playlists &
  "\p\ttags: " & $db.root.tags &
  "\p\talbumsToMediaFiles: " & $db.root.albumsToMediaFiles &
  "\p\tartistsToMediaFiles: " & $db.root.artistsToMediaFiles &
  "\p\tartistsAndAlbums: " & $db.root.artistsAndAlbums &
  "\p\tplaylistsToMediaFiles: " & $db.root.playlistsToMediaFiles &
  "\p\ttagsAndMediaFiles: " & $db.root.tagsAndMediaFiles &
  "\p\tmediaFileHashToId: " & $db.root.mediaFileHashToId & "\p"

## ========================================================================= ##
##                            API Contract                                   ##
## ========================================================================= ##
proc initDbRoot(): DbRoot;
proc initDb*(path: string): WdiwtltDb;
proc loadDb*(path: string): WdiwtltDb;
proc persist*(db: WdiwtltDb): void;

## Albums
## --------------------
proc add*(db: WdiwtltDb, a: Album);
proc remove*(db: WdiwtltDb, a: Album);
proc update*(db: WdiwtltDb, a: Album);

proc add*(db: WdiwtltDb, a: Album, mf: MediaFile);
proc remove*(db: WdiwtltDb, a: Album, mf: MediaFile);

proc allAlbums*(db: WdiwtltDb): seq[Album];

proc findAlbumById*(db: WdiwtltDb, id: UUID): Option[Album];
proc findAlbumsByArtist*(db: WdiwtltDb, a: Artist): seq[Album];
proc findAlbumsByMediaFile*(db: WdiwtltDb, mf: MediaFile): seq[Album];
proc findAlbumsByName*(db: WdiwtltDb, name: string): seq[Album];

proc findAllRelatedAlbums*(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[Album];

## Artists
## --------------------
proc add*(db: WdiwtltDb, a: Artist);
proc remove*(db: WdiwtltDb, a: Artist);
proc update*(db: WdiwtltDb, a: Artist);

proc add*(db: WdiwtltDb, a: Artist, mf: MediaFile);
proc remove*(db: WdiwtltDb, a: Artist, mf: MediaFile);

proc associate*(db: WdiwtltDb, artist: Artist, album: Album);
proc disassociate*(db: WdiwtltDb, artist: Artist, album: Album);
proc isAssociated*(db: WdiwtltDb, artist: Artist, album: Album): bool;

proc allArtists*(db: WdiwtltDb): seq[Artist];

proc findArtistById*(db: WdiwtltDb, id: UUID): Option[Artist];
proc findArtistsByAlbum*(db: WdiwtltDb, a: Album): seq[Artist];
proc findArtistsByMediaFile*(db: WdiwtltDb, mf: MediaFile): seq[Artist];
proc findArtistsByName*(db: WdiwtltDb, name: string): seq[Artist];

proc findAllRelatedArtists*(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[Artist];

## Bookmarks
## --------------------
proc add*(db: WdiwtltDb, b: Bookmark);
proc update*(db: WdiwtltDb, b: Bookmark);
proc remove*(db: WdiwtltDb, b: Bookmark);

proc allBookmarks*(db: WdiwtltDb): seq[Bookmark];

proc findBookmarkById*(db: WdiwtltDb, id: UUID): Option[Bookmark];
proc findBookmarksByPlaylist*(db: WdiwtltDb, p: Playlist): seq[Bookmark];
proc findBookmarksByName*(db: WdiwtltDb, name: string): seq[Bookmark];

proc findAllRelatedBookmarks*(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[Bookmark];

## Media Files
## --------------------
proc add*(db: WdiwtltDb, mf: MediaFile);
proc remove*(db: WdiwtltDb, mf: MediaFile);
proc update*(db: WdiwtltDb, mf: MediaFile);

proc allMediaFiles*(db: WdiwtltDb): seq[MediaFile];

proc findMediaFileByHash*(db: WdiwtltDb, hash: string): Option[MediaFile];
proc findMediaFileById*(db: WdiwtltDb, id: UUID): Option[MediaFile];
proc findMediaFileByPath*(db: WdiwtltDb, path: string): Option[MediaFile];
proc findMediaFilesByAlbum*(db: WdiwtltDb, a: Album): seq[MediaFile];
proc findMediaFilesByArtist*(db: WdiwtltDb, a: Artist): seq[MediaFile];
proc findMediaFilesByBookmark*(db: WdiwtltDb, b: Bookmark): seq[MediaFile];
proc findMediaFilesByName*(db: WdiwtltDb, name: string): seq[MediaFile];
proc findMediaFilesByPlaylist*(db: WdiwtltDb, p: Playlist): seq[MediaFile];
proc findMediaFilesByTag*(db: WdiwtltDb, tag: Tag): seq[MediaFile];

proc findAllRelatedMediaFiles*(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[MediaFile];

## Playlists
## --------------------
proc add*(db: WdiwtltDb, p: Playlist);
proc remove*(db: WdiwtltDb, p: Playlist);
proc update*(db: WdiwtltDb, p: Playlist);

proc allPlaylists*(db: WdiwtltDb): seq[Playlist];

proc findPlaylistById*(db: WdiwtltDb, id: UUID): Option[Playlist];
proc findPlaylistsByMediaFile*(db: WdiwtltDb, mf: MediaFile): seq[Playlist];
proc findPlaylistsByName*(db: WdiwtltDb, name: string): seq[Playlist];

proc findAllRelatedPlaylists*(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[Playlist];

## Tags
## --------------------
proc add*(db: WdiwtltDb, t: Tag);
proc remove*(db: WdiwtltDb, t: Tag);
proc update*(db: WdiwtltDb, t: Tag);

proc allTags*(db: WdiwtltDb): seq[Tag];

proc findTagById*(db: WdiwtltDb, id: UUID): Option[Tag];
proc findTagByName*(db: WdiwtltDb, name: string): Option[Tag];
proc findTagsByMediaFile*(db: WdiwtltDb, mf: MediaFile): seq[Tag];

proc findAllRelatedTags*(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[Tag];

## General
## --------------------

proc getAll*(db: WdiwtltDb, mt: ModelType): seq[BaseModel];

proc findAllRelatedModels*(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel],
    targetModelType: ModelType
  ): seq[BaseModel];

## Housekeeping
## --------------------
proc pruneStalePlaylists*(db: WdiwtltDb, ts: DateTime);
proc pruneStaleBookmarks*(db: WdiwtltDb, ts: DateTime);
proc removeEmptyAlbums*(db: WdiwtltDb): void;
proc removeEmptyArtists*(db: WdiwtltDb): void;
proc removeEmptyPlaylists*(db: WdiwtltDb): void;

## ========================================================================= ##
##                            API Implementation                             ##
## ========================================================================= ##

## To JSON
## --------------------
proc `%`(dt: DateTime): JsonNode = %(dt.formatIso8601)
proc `%`(u: UUID): JsonNode = %($u)
proc `%`(table: TableRef): JsonNode =
  result = newJObject()
  for k, v in table.pairs: result[$k] = %v

proc `%`(tup: tuple[albumId, mediaFileId: UUID]): JsonNode =
  result = newJObject()
  result["albumId"] = %($tup.albumId)
  result["mediaFileId"] = %($tup.mediaFileId)

proc `%`(tup: tuple[artistId, mediaFileId: UUID]): JsonNode =
  result = newJObject()
  result["artistId"] = %($tup.artistId)
  result["mediaFileId"] = %($tup.mediaFileId)

proc `%`(tup: tuple[artistId, albumId: UUID]): JsonNode =
  result = newJObject()
  result["artistId"] = %($tup.artistId)
  result["albumId"] = %($tup.albumId)

proc `%`(tup: tuple[tagId: UUID, mediaFileId: UUID]): JsonNode =
  result = newJObject()
  result["tagId"] = %($tup.tagId)
  result["mediaFileId"] = %($tup.mediaFileId)

proc `%`(root: DbRoot): JsonNode =
  %*{
    "albums": root.albums,
    "artists": root.artists,
    "bookmarks": root.bookmarks,
    "mediaFiles": root.mediaFiles,
    "playlists": root.playlists,
    "tags": root.tags,
    "albumsToMediaFiles": root.albumsToMediaFiles,
    "artistsToMediaFiles": root.artistsToMediaFiles,
    "artistsAndAlbums": root.artistsAndAlbums,
    "playlistsToMediaFiles": root.playlistsToMediaFiles,
    "tagsAndMediaFiles": root.tagsAndMediaFiles,
    "mediaFileHashToId": root.mediaFileHashToId
  }

## From JSON
## --------------------
proc parseAlbumsTable(n: JsonNode): TableRef[UUID, Album] =
  result = newTable[UUID, Album]()
  for strId in n.keys: result[parseUUID(strId)] = parseAlbum(n[strId])

proc parseArtistsTable(n: JsonNode): TableRef[UUID, Artist] =
  result = newTable[UUID, Artist]()
  for strId in n.keys: result[parseUUID(strId)] = parseArtist(n[strId])

proc parseBookmarksTable(n: JsonNode): TableRef[UUID, Bookmark] =
  result = newTable[UUID, Bookmark]()
  for strId in n.keys: result[parseUUID(strId)] = parseBookmark(n[strId])

proc parseMediaFilesTable(n: JsonNode): TableRef[UUID, MediaFile] =
  result = newTable[UUID, MediaFile]()
  for strId in n.keys: result[parseUUID(strId)] = parseMediaFile(n[strId])

proc parsePlaylistsTable(n: JsonNode): TableRef[UUID, Playlist] =
  result = newTable[UUID, Playlist]()
  for strId in n.keys: result[parseUUID(strId)] = parsePlaylist(n[strId])

proc parseTagsTable(n: JsonNode): TableRef[UUID, Tag] =
  result = newTable[UUID, Tag]()
  for strId in n.keys: result[parseUUID(strId)] = parseTag(n[strId])

proc parseMediaFilesAssociationTable(n: JsonNode): TableRef[UUID, seq[UUID]] =
  result = newTable[UUID, seq[UUID]]()
  for strId in n.keys:
    result[parseUUID(strId)] = n[strId].getElems --> map(parseUUID(it.getStr))

proc parseArtistsAndAlbumsList(n: JsonNode):
    seq[tuple[artistId, albumId: UUID]] =
  n.getElems --> map((it.parseUUID("artistId"), it.parseUUID("albumId")))

proc parseTagsAndMediaFilesList(n: JsonNode):
  seq[tuple[tagId: UUID, mediaFileId: UUID]] =
  n.getElems --> map((it.parseUUID("tagId"), it.parseUUID("mediaFileId")))

proc parseMediaFileHashToIdTable(n: JsonNode): TableRef[string, UUID] =
  result = newTable[string, UUID]()
  for hash in n.keys: result[hash] = parseUUID(n[hash].getStr)

proc parseDbRoot(n: JsonNode): DbRoot =
  let tagsTable = n.getOrFail("tags").parseTagsTable
  result = DbRoot(
    albums: n.getOrFail("albums").parseAlbumsTable,
    artists: n.getOrFail("artists").parseArtistsTable,
    bookmarks: n.getOrFail("bookmarks").parseBookmarksTable,
    mediaFiles: n.getOrFail("mediaFiles").parseMediaFilesTable,
    playlists: n.getOrFail("playlists").parsePlaylistsTable,
    tags: tagsTable,

    albumsToMediaFiles:
      n.getOrFail("albumsToMediaFiles").parseMediaFilesAssociationTable,
    artistsToMediaFiles:
      n.getOrFail("artistsToMediaFiles").parseMediaFilesAssociationTable,
    artistsAndAlbums:
      n.getOrFail("artistsAndAlbums").parseArtistsAndAlbumsList,
    playlistsToMediaFiles:
      n.getOrFail("playlistsToMediaFiles").parseMediaFilesAssociationTable,
    tagsAndMediaFiles:
      n.getOrFail("tagsAndMediaFiles").parseTagsAndMediaFilesList,
    mediaFileHashToId:
      n.getOrFail("mediaFileHashToId").parseMediaFileHashToIdTable
  )

  result.tagSet = toHashSet(tagsTable.values.toSeq --> map(it.name))


## Internals
## --------------------
proc initDbRoot(): DbRoot =
  DbRoot(
    albums: newTable[UUID, Album](),
    artists: newTable[UUID, Artist](),
    bookmarks: newTable[UUID, Bookmark](),
    mediaFiles: newTable[UUID, MediaFile](),
    playlists: newTable[UUID, Playlist](),
    tags: newTable[UUID, Tag](),
    albumsToMediaFiles: newTable[UUID, seq[UUID]](),
    artistsToMediaFiles: newTable[UUID, seq[UUID]](),
    artistsAndAlbums: @[],
    playlistsToMediaFiles: newTable[UUID, seq[UUID]](),
    tagsAndMediaFiles: @[],
    tagSet: initHashSet[string](),
    mediaFileHashToId: newTable[string, UUID]())

proc initDb*(path: string): WdiwtltDb =
  WdiwtltDb(
    jsonFilePath: path,
    root: initDbRoot())

proc loadDb*(path: string): WdiwtltDb =
  if not fileExists(path):
    raise newException(Exception, "Unable to open database file '" & path & "'")

  result = WdiwtltDb(
    jsonFilePath: path,
    root: parseDbRoot(parseJson(path.readFile)))
  debug "loaded DB"

proc persist*(db: WdiwtltDb): void =
  db.jsonFilePath.writeFile($(%db.root))

## Albums
## --------------------
proc add*(db: WdiwtltDb, a: Album) = db.root.albums[a.id] = a

proc remove*(db: WdiwtltDb, a: Album) =
  db.root.albums.del(a.id)
  db.root.albumsToMediaFiles.del(a.id)
  db.root.artistsAndAlbums =
    db.root.artistsAndAlbums --> filter(it.albumId != a.id)

proc update*(db: WdiwtltDb, a: Album) = db.add(a)

proc add*(db: WdiwtltDb, a: Album, mf: MediaFile) =
  if not db.root.albumsToMediaFiles.contains(a.id):
    db.root.albumsToMediaFiles[a.id] = @[mf.id]
  else:
    let idsInAlbum = db.root.albumsToMediaFiles[a.id]
    if not idsInAlbum.contains(mf.id):
      db.root.albumsToMediaFiles[a.id].add(mf.id)

proc remove*(db: WdiwtltDb, a: Album, mf: MediaFile) =
  if db.root.albumsToMediaFiles.contains(a.id):
    db.root.albumsToMediaFiles[a.id] =
      db.root.albumsToMediaFiles[a.id] --> filter(it != mf.id)

proc allAlbums*(db: WdiwtltDb): seq[Album] = db.root.albums.values.toSeq

proc findAlbumById*(db: WdiwtltDb, id: UUID): Option[Album] =
  if not db.root.albums.contains(id): return none[Album]()
  return some(db.root.albums[id])

proc findAlbumsByArtist*(db: WdiwtltDb, a: Artist): seq[Album] =
  db.root.artistsAndAlbums -->
    filter(it.artistId == a.id and db.root.artists.contains(a.id)).
    map(db.root.albums[it.albumId])

proc findAlbumsByMediaFile*(db: WdiwtltDb, mf: MediaFile): seq[Album] =
  db.root.albumsToMediaFiles.pairs.toSeq -->
    filter(db.root.albums.contains(it[0]) and it[1].contains(mf.id)).
    map(db.root.albums[it[0]])

proc findAlbumsByName*(db: WdiwtltDb, name: string): seq[Album] =
  db.root.albums.values.toSeq --> filter(it.name == name)

proc findAllRelatedAlbums(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[Album] =

  if sourceModels.len == 0: return @[]
  let firstModel = sourceModels[0]

  if firstModel of Album: return sourceModels --> map(it.Album)

  elif firstModel of Artist:
    return sourceModels -->
      map(db.findAlbumsByArtist(it.Artist)).flatten()

  elif firstModel of Bookmark or
       firstModel of Playlist or
       firstModel of MediaFile or
       firstModel of Tag:

    return deduplicate(
      db.findAllRelatedMediaFiles(sourceModels) -->
        map(db.findAlbumsByMediaFile(it)).flatten())

  else:
    warn "unknown source model type, returning empty selection"
    debug "findAllRelatedAlbums: first source model was " & $firstModel

## Artists
## --------------------
proc add*(db: WdiwtltDb, a: Artist) = db.root.artists[a.id] = a

proc remove*(db: WdiwtltDb, a: Artist) =
  db.root.artists.del(a.id)
  db.root.artistsToMediaFiles.del(a.id)
  db.root.artistsAndAlbums =
    db.root.artistsAndAlbums --> filter(it.artistId != a.id)

proc update*(db: WdiwtltDb, a: Artist) = db.add(a)

proc allArtists*(db: WdiwtltDb): seq[Artist] = db.root.artists.values.toSeq

proc findArtistById*(db: WdiwtltDb, id: UUID): Option[Artist] =
  if not db.root.artists.contains(id): return none[Artist]()
  return some(db.root.artists[id])

proc findArtistsByAlbum*(db: WdiwtltDb, a: Album): seq[Artist] =
  db.root.artistsAndAlbums -->
    filter(it.albumId == a.id and db.root.albums.contains(a.id)).
    map(db.root.artists[it.artistId])

proc findArtistsByMediaFile*(db: WdiwtltDb, mf: MediaFile): seq[Artist] =
  db.root.artistsToMediaFiles.pairs.toSeq -->
    filter(db.root.artists.contains(it[0]) and it[1].contains(mf.id)).
    map(db.root.artists[it[0]])

proc findArtistsByName*(db: WdiwtltDb, name: string): seq[Artist] =
  db.root.artists.values.toSeq --> filter(it.name == name)

proc findAllRelatedArtists(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[Artist] =

  if sourceModels.len == 0: return @[]
  let firstModel = sourceModels[0]

  if firstModel of Album:
    return deduplicate(
      sourceModels -->
        map(db.findArtistsByAlbum(it.Album)).flatten())

  elif firstModel of Artist: return sourceModels --> map(it.Artist)

  elif firstModel of Bookmark or
       firstModel of Playlist or
       firstModel of MediaFile or
       firstModel of Tag:

    return deduplicate(
      db.findAllRelatedMediaFiles(sourceModels) -->
        map(db.findArtistsByMediaFile(it)).flatten())

  else:
    warn "unknown source model type, returning empty selection"
    debug "findAllRelatedArtists: first source model was " & $firstModel

proc add*(db: WdiwtltDb, a: Artist, mf: MediaFile) =
  if not db.root.artistsToMediaFiles.contains(a.id):
    db.root.artistsToMediaFiles[a.id] = @[mf.id]
  else:
    let idsInArtist = db.root.artistsToMediaFiles[a.id]
    if not idsInArtist.contains(mf.id):
      db.root.artistsToMediaFiles[a.id].add(mf.id)

proc remove*(db: WdiwtltDb, a: Artist, mf: MediaFile) =
  if db.root.artistsToMediaFiles.contains(a.id):
    db.root.artistsToMediaFiles[a.id] =
      db.root.artistsToMediaFiles[a.id] --> filter(it != mf.id)

proc isAssociated*(
    db: WdiwtltDb,
    artist: Artist,
    album: Album
  ): bool =
  let matching = db.root.artistsAndAlbums -->
    filter(it.artistId == artist.id and it.albumId == album.id)
  return matching.len > 0

proc associate*(db: WdiwtltDb, artist: Artist, album: Album) =
  if not db.isAssociated(artist, album):
    db.root.artistsAndAlbums.add((artist.id, album.id))

proc disassociate*(db: WdiwtltDb, artist: Artist, album: Album) =
  db.root.artistsAndAlbums = db.root.artistsAndAlbums -->
    filter(it.albumId != album.id or it.artistId != artist.id)

## Bookmarks
## --------------------
proc add*(db: WdiwtltDb, b: Bookmark) = db.root.bookmarks[b.id] = b
proc remove*(db: WdiwtltDb, b: Bookmark) = db.root.bookmarks.del(b.id)
proc update*(db: WdiwtltDb, b: Bookmark) = db.add(b)

proc allBookmarks*(db: WdiwtltDb): seq[Bookmark] =
  db.root.bookmarks.values.toSeq

proc findBookmarkById*(db: WdiwtltDb, id: UUID): Option[Bookmark] =
  if not db.root.bookmarks.contains(id): return none[Bookmark]()
  return some(db.root.bookmarks[id])

proc findBookmarksByName*(db: WdiwtltDb, name: string): seq[Bookmark] =
  db.root.bookmarks.values.toSeq --> filter(it.name == name)

proc findBookmarksByPlaylist*(db: WdiwtltDb, p: Playlist): seq[Bookmark] =
  db.root.bookmarks.values.toSeq -->
    filter(it.playlistId == p.id)

proc findAllRelatedBookmarks*(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[Bookmark] =

  if sourceModels.len == 0: return @[]
  let firstModel = sourceModels[0]

  if firstModel of Album:
    return deduplicate(
      sourceModels -->
        map(db.findMediaFilesByAlbum(it.Album)).flatten().
        map(db.findPlaylistsByMediaFile(it)).flatten().
        map(db.findBookmarksByPlaylist(it)).flatten())

  elif firstModel of Artist:
    return deduplicate(
      sourceModels -->
        map(db.findMediaFilesByArtist(it.Artist)).flatten().
        map(db.findPlaylistsByMediaFile(it)).flatten().
        map(db.findBookmarksByPlaylist(it)).flatten())

  elif firstModel of Bookmark:
    return sourceModels --> map(it.Bookmark)

  elif firstModel of MediaFile:
    return deduplicate(
      sourceModels -->
        map(db.findPlaylistsByMediaFile(it.MediaFile)).flatten().
        map(db.findBookmarksByPlaylist(it)).flatten())

  elif firstModel of Playlist:
    return deduplicate(
      sourceModels --> map(db.findBookmarksByPlaylist(it.Playlist)).flatten())

  elif firstModel of Tag:
    return deduplicate(
      sourceModels -->
        map(db.findMediaFilesByTag(it.Tag)).flatten().
        map(db.findPlaylistsByMediaFile(it)).flatten().
        map(db.findBookmarksByPlaylist(it)).flatten())

  else:
    warn "unknown source model type, returning empty selection"
    debug "findAllRelatedBookmarks: first source model was " & $firstModel

## Media Files
## --------------------
proc add*(db: WdiwtltDb, mf: MediaFile) =
  db.root.mediaFiles[mf.id] = mf
  db.root.mediaFileHashToId[mf.fileHash] = mf.id

proc remove*(db: WdiwtltDb, mf: MediaFile) =
  db.root.mediaFiles.del(mf.id)

  for albumId, mfIds in db.root.albumsToMediaFiles.pairs:
    if mfIds.contains(mf.id):
      db.remove(db.root.albums[albumId], mf)

  db.root.mediaFileHashToId.del(mf.fileHash)

proc update*(db: WdiwtltDb, mf: MediaFile) = db.add(mf)

proc allMediaFiles*(db: WdiwtltDb): seq[MediaFile] =
  db.root.mediaFiles.values.toSeq

proc findMediaFileByHash*(db: WdiwtltDb, hash: string): Option[MediaFile] =
  if db.root.mediaFileHashToId.contains(hash):
    let mfId = db.root.mediaFileHashToId[hash]
    if db.root.mediaFiles.contains(mfId):
      return some(db.root.mediaFiles[mfId])

  return none[MediaFile]()

proc findMediaFileById*(db: WdiwtltDb, id: UUID): Option[MediaFile] =
  if not db.root.mediaFiles.contains(id): return none[MediaFile]()
  return some(db.root.mediaFiles[id])

proc findMediaFileByPath*(db: WdiwtltDb, path: string): Option[MediaFile] =
  for mf in db.root.mediaFiles.values:
    if mf.filePath == path: return some(mf)
  return none[MediaFile]()

proc findMediaFilesByAlbum*(db: WdiwtltDb, a: Album): seq[MediaFile] =
  if db.root.albumsToMediaFiles.contains(a.id):
    return db.root.albumsToMediaFiles[a.id] -->
      filter(db.root.mediaFiles.contains(it)).
      map(db.root.mediaFiles[it])

proc findMediaFilesByArtist*(db: WdiwtltDb, a: Artist): seq[MediaFile] =
  if db.root.artistsToMediaFiles.contains(a.id):
    return db.root.artistsToMediaFiles[a.id] -->
      filter(db.root.mediaFiles.contains(it)).
      map(db.root.mediaFiles[it])

proc findMediaFilesByBookmark*(db: WdiwtltDb, b: Bookmark): seq[MediaFile] =
  if db.root.playlists.contains(b.playlistId):
    return db.findMediaFilesByPlaylist(db.root.playlists[b.playlistId])

proc findMediaFilesByName*(db: WdiwtltDb, name: string): seq[MediaFile] =
  db.root.mediaFiles.values.toSeq --> filter(it.name == name)

proc findMediaFilesByPlaylist*(db: WdiwtltDb, p: Playlist): seq[MediaFile] =
  if db.root.playlistsToMediaFiles.contains(p.id):
    return db.root.playlistsToMediaFiles[p.id] -->
      filter(db.root.mediaFiles.contains(it)).
      map(db.root.mediaFiles[it])

proc findMediaFilesByTag*(db: WdiwtltDb, tag: Tag): seq[MediaFile] =
  return db.root.tagsAndMediaFiles -->
    filter(db.root.mediaFiles.contains(it.mediaFileId) and it.tagId == tag.id).
    map(db.root.mediaFiles[it.mediaFileId])

proc findAllRelatedMediaFiles(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[MediaFile] =
  # Note, contrary to other model types, we do not deduplicate the media file
  # lookup  as the expected behavior when listing by multiple
  # albums/artists/etc. is to get all the media files for the model in order.
  # In general there should be duplication unless the exact same media file is
  # present in multiple artist/album/etc. collections. This would need to be
  # done intentionally, as the current scan implementation has no mechanism for
  # creating this kind of duplicate link (being path-based).

  if sourceModels.len == 0: return @[]
  let firstModel = sourceModels[0]

  if firstModel of Album:
    return sourceModels -->
      map(db.findMediaFilesByAlbum(it.Album)).flatten()

  elif firstModel of Artist:
    return sourceModels -->
      map(db.findMediaFilesByArtist(it.Artist)).flatten()

  elif firstModel of Bookmark:
    return sourceModels -->
      map(db.findMediaFilesByBookmark(it.Bookmark)).flatten()

  elif firstModel of MediaFile:
    return sourceModels --> map(it.MediaFile)

  elif firstModel of Playlist:
    return sourceModels -->
      map(db.findMediaFilesByPlaylist(it.Playlist)).flatten()

  elif firstModel of Tag:
    return sourceModels -->
      map(db.findMediaFilesByTag(it.Tag)).flatten()

  else:
    warn "unknown source model type, returning empty selection"
    debug "findAllRelatedMediaFiles: first source model was " & $firstModel

## Playlists
## --------------------
proc add*(db: WdiwtltDb, p: Playlist) = db.root.playlists[p.id] = p

proc remove*(db: WdiwtltDb, p: Playlist) =
  db.root.playlists.del(p.id)
  db.root.playlistsToMediaFiles.del(p.id)
  for b in db.root.bookmarks.values:
    if b.playlistId == p.id: db.remove(b)

proc update*(db: WdiwtltDb, p: Playlist) = db.add(p)

proc allPlaylists*(db: WdiwtltDb): seq[Playlist] =
  db.root.playlists.values.toSeq

proc findPlaylistById*(db: WdiwtltDb, id: UUID): Option[Playlist] =
  if not db.root.playlists.contains(id): return none[Playlist]()
  return some(db.root.playlists[id])

proc findPlaylistsByMediaFile*(db: WdiwtltDb, mf: MediaFile): seq[Playlist] =
  db.root.playlistsToMediaFiles.pairs.toSeq -->
    filter(db.root.playlists.contains(it[0]) and it[1].contains(mf.id)).
    map(db.root.playlists[it[0]])

proc findPlaylistsByName*(db: WdiwtltDb, name: string): seq[Playlist] =
  db.root.playlists.values.toSeq --> filter(it.name == name)

proc findAllRelatedPlaylists(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[Playlist] =

  if sourceModels.len == 0: return @[]
  let firstModel = sourceModels[0]

  if firstModel of Bookmark:
    return sourceModels -->
      map(it.Bookmark).
      filter(db.root.playlists.contains(it.playlistId)).
      map(db.root.playlists[it.playlistId])

  elif firstModel of Playlist:
    return sourceModels --> map(it.Playlist)

  elif firstModel of Album or
     firstModel of Artist or
     firstModel of MediaFile or
     firstModel of Tag:

    return deduplicate(
      db.findAllRelatedMediaFiles(sourceModels) -->
        map(db.findPlaylistsByMediaFile(it)).flatten())

  else:
    warn "unknown source model type, returning empty selection"
    debug "findAllRelatedPlaylists: first source model was " & $firstModel

## Tags
## --------------------
proc add*(db: WdiwtltDb, t: Tag) =
  db.root.tags[t.id] = t
  db.root.tagSet.incl(t.name)

proc remove*(db: WdiwtltDb, t: Tag) =
  db.root.tags.del(t.id)
  db.root.tagSet.excl(t.name)
  db.root.tagsAndMediaFiles = db.root.tagsAndMediaFiles -->
    filter(it.tagId != t.id)

proc update*(db: WdiwtltDb, t: Tag) =
  if db.root.tags.contains(t.id):
    db.root.tagSet.excl(db.root.tags[t.id].name)

  db.add(t)

proc allTags*(db: WdiwtltDb): seq[Tag] =
  db.root.tags.values.toSeq

proc findTagById*(db: WdiwtltDb, id: UUID): Option[Tag] =
  if not db.root.tags.contains(id): return none[Tag]()
  return some(db.root.tags[id])

proc findTagByName*(db: WdiwtltDb, name: string): Option[Tag] =
  if not db.root.tagSet.contains(name): return none[Tag]()
  let foundTags = db.root.tags.values.toSeq --> filter(it.name == name)
  if foundTags.len == 0: return none[Tag]()
  else: return some(foundTags[0])

proc findTagsByMediaFile*(db: WdiwtltDb, mf: MediaFile): seq[Tag] =
  db.root.tagsAndMediaFiles -->
    filter(it.mediaFileId == mf.id).
    map(db.root.tags[it.tagId])

proc findAllRelatedTags(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel]
  ): seq[Tag] =

  if sourceModels.len == 0: return @[]
  let firstModel = sourceModels[0]

  if firstModel of Tag:
    return sourceModels --> map(it.Tag)

  elif firstModel of Album or
     firstModel of Artist or
     firstModel of Bookmark or
     firstModel of MediaFile or
     firstModel of Playlist:

    return deduplicate(
      db.findAllRelatedMediaFiles(sourceModels) -->
        map(db.findTagsByMediaFile(it)).flatten())

  else:
    warn "unknown source model type, returning empty selection"

## General
## --------------------

proc getAll*(db: WdiwtltDb, mt: ModelType): seq[BaseModel] =
  case mt:
  of mtAlbum: db.root.albums.values.toSeq --> map(it.BaseModel)
  of mtArtist: db.root.artists.values.toSeq --> map(it.BaseModel)
  of mtBookmark: db.root.bookmarks.values.toSeq --> map(it.BaseModel)
  of mtMediaFile: db.root.bookmarks.values.toSeq --> map(it.BaseModel)
  of mtPlaylist: db.root.playlists.values.toSeq --> map(it.BaseModel)
  of mtTag: db.root.tags.values.toSeq --> map(it.BaseModel)


proc findAllRelatedModels*(
    db: WdiwtltDb,
    sourceModels: seq[BaseModel],
    targetModelType: ModelType
  ): seq[BaseModel] =

  if sourceModels.len <= 0: return @[]

  case targetModelType:
  of mtAlbum:
    return db.findAllRelatedAlbums(sourceModels) --> map(it.BaseModel)

  of mtArtist:
    return db.findAllRelatedArtists(sourceModels) --> map(it.BaseModel)

  of mtBookmark:
    return db.findAllRelatedBookmarks(sourceModels) --> map(it.BaseModel)

  of mtMediaFile:
    return db.findAllRelatedMediaFiles(sourceModels) --> map(it.BaseModel)

  of mtPlaylist:
    return db.findAllRelatedPlaylists(sourceModels) --> map(it.BaseModel)

  of mtTag:
    return db.findAllRelatedTags(sourceModels) --> map(it.BaseModel)

## Housekeeping
## --------------------
proc pruneStalePlaylists*(db: WdiwtltDb, ts: DateTime) =
  db.root.playlists.values -->
    filter(it.lastUsed > ts).foreach(db.remove(it))

proc pruneStaleBookmarks*(db: WdiwtltDb, ts: DateTime) =
  db.root.bookmarks.values -->
    filter(it.lastUsed > ts).foreach(db.remove(it))

proc removeEmptyAlbums*(db: WdiwtltDb): void =
  let emptyAlbumIds = db.root.albumsToMediaFiles.pairs -->
    filter(it[1].len == 0).map(it[0])

  for i in emptyAlbumIds:
    db.root.albums.del(i)
    db.root.albumsToMediaFiles.del(i)
    db.root.artistsAndAlbums = db.root.artistsAndAlbums -->
      filter(it.albumId != i)

proc removeEmptyArtists*(db: WdiwtltDb): void =
  let emptyArtistIds = db.root.artistsToMediaFiles.pairs -->
    filter(it[1].len == 0).map(it[0])

  for i in emptyArtistIds:
    db.root.artists.del(i)
    db.root.artistsToMediaFiles.del(i)
    db.root.artistsAndAlbums = db.root.artistsAndAlbums -->
      filter(it.artistId != i)


proc removeEmptyPlaylists*(db: WdiwtltDb): void =
  let emptyPlaylistIds = db.root.playlistsToMediaFiles.pairs -->
    filter(it[1].len == 0).map(it[0])

  for i in emptyPlaylistIds:
    db.root.playlists.del(i)
    db.root.playlistsToMediaFiles.del(i)
    let emptyBookmarkIds = db.root.bookmarks.pairs -->
      filter(it[1].playlistId == i).map(it[0])
    for j in emptyBookmarkIds: db.root.bookmarks.del(j)
