import std/[logging, strutils, tables, terminal]
import ./cliconstants

type
  TextStyle* = enum
    tsAlbum, tsArtist, tsCommand, tsError, tsFile, tsNormal, tsOption,
    tsPlaylist, tsPrompt, tsStatus, tsTitle,

  StyledText* = object
    style: TextStyle
    text: string

let styleToColorMap = newTable[TextStyle, ForegroundColor]([
  (tsAlbum, fgBlue),
  (tsArtist, fgRed),
  (tsCommand, fgYellow),
  (tsError, fgRed),
  (tsFile, fgGreen),
  (tsNormal, fgWhite),
  (tsOption, fgGreen),
  (tsPlaylist, fgGreen),
  (tsPrompt, fgYellow),
  (tsStatus, fgCyan),
  (tsTitle, fgWhite)
])

let styleToFormatMap = newTable[TextStyle, set[Style]]([
  (tsAlbum, {}),
  (tsArtist, {}),
  (tsCommand, {styleBright}),
  (tsError, {styleBright}),
  (tsFile, {}),
  (tsNormal, {}),
  (tsOption, {styleBright}),
  (tsPlaylist, {}),
  (tsPrompt, {styleBright}),
  (tsStatus, {}),
  (tsTitle, {styleBright})
])

func parseStyledText*(text: string): seq[StyledText] =
  result = newseq[StyledText]()
  var curIdx = 0
  var curStyle = tsNormal

  while curIdx < text.len:
    let nextMarkerIdx = text.find(FMT_MARK, curIdx)

    if nextMarkerIdx == -1:
      result.add(StyledText(
        style: curStyle,
        text: text[curIdx..^1]))
      curIdx = text.len

    else:
      let endMarkerIdx = text.find(FMT_MARK, nextMarkerIdx + FMT_MARK.len)
      if endMarkerIdx == -1:
        raise newException(Exception,
          "invalid text formatting, missing ending marker")

      result.add(StyledText(
        style: curStyle,
        text: text[curIdx..<nextMarkerIdx]))

      let styleName = text[(nextMarkerIdx+FMT_MARK.len)..<endMarkerIdx]
      case styleName:
      of "NORMAL": curStyle = tsNormal
      of "COMMAND": curStyle = tsCommand
      of "ALBUM": curStyle = tsAlbum
      of "ARTIST": curStyle = tsArtist
      of "ERROR": curStyle = tsError
      of "FILE": curStyle = tsFile
      of "OPTION": curStyle = tsOption
      of "PLAYLIST": curStyle = tsPlaylist
      of "PROMPT": curStyle = tsPrompt
      of "STATUS": curStyle = tsStatus
      of "TITLE": curStyle = tsTitle
      else: curStyle = tsNormal

      curIdx = endMarkerIdx + FMT_MARK.len

proc writeStyledText*(file: File, text: string) =
  let parsed = parseStyledText(text)

  for p in parsed:
    if styleToColorMap.contains(p.style) and
       styleToFormatMap.contains(p.style):

      file.styledWrite(
        styleToColorMap[p.style], styleToFormatMap[p.style], p.text)

    else:
      warn "Missing text styling for the " & $p.style & " style."
      file.styledWrite(p.text)


proc writeStyledTextLine*(file: File, text: string) =
  file.writeStyledText(text)
  file.writeLine ""
